import { isArray } from 'lodash';
import { from, Subject } from 'rxjs';
import { forkJoin } from 'rxjs';
import { FormGroup } from './form-group';
export const [VALID, INVALID, PENDING, DISABLED] = ['VALID', 'INVALID', 'PENDING', 'DISABLED'];
function find(control, path, delimiter) {
    if (path == null)
        return null;
    if (!Array.isArray(path)) {
        path = path.split(delimiter);
    }
    if (Array.isArray(path) && path.length === 0)
        return null;
    // Not using Array.reduce here due to a Chrome 80 bug
    // https://bugs.chromium.org/p/chromium/issues/detail?id=1049982
    let controlToFind = control;
    path.forEach((name) => {
        if (controlToFind instanceof FormGroup) {
            controlToFind = name in controlToFind.controls ? controlToFind.controls[name] : null;
        }
        else {
            controlToFind = null;
        }
    });
    return controlToFind;
}
function mergeErrors(arrayOfErrors) {
    const res = {};
    arrayOfErrors.forEach((errors) => {
        if (errors != null) {
            Object.assign(res, errors);
        }
    });
    return Object.keys(res).length === 0 ? null : res;
}
function composeValidators(validators) {
    const presentValidators = validators.filter((validator) => validator !== null);
    if (presentValidators.length === 0) {
        return null;
    }
    return function (control) {
        return mergeErrors(presentValidators.map((fn) => fn(control)));
    };
}
function composeAsyncValidators(validators) {
    const presentValidators = validators.filter((validator) => validator !== null);
    if (presentValidators.length === 0) {
        return null;
    }
    return function (control) {
        return new Promise((resolve) => {
            const observables = forkJoin(presentValidators.map((fn) => fn(control)));
            observables.subscribe({
                next: (errors) => {
                    resolve(mergeErrors(errors));
                },
            });
        });
    };
}
function coerceToValidator(validator) {
    return isArray(validator) ? composeValidators(validator) : validator;
}
function coerceToAsyncValidator(asyncValidator) {
    return isArray(asyncValidator) ? composeAsyncValidators(asyncValidator) : asyncValidator;
}
function makeValidatorsArray(validators) {
    if (!validators) {
        return [];
    }
    return isArray(validators) ? validators : [validators];
}
function hasValidator(validators, validator) {
    return Array.isArray(validators) ? validators.includes(validator) : validators === validator;
}
function addValidators(validators, currentValidators) {
    const current = makeValidatorsArray(currentValidators);
    const validatorsToAdd = makeValidatorsArray(validators);
    return current.concat(validatorsToAdd.filter((validator) => !hasValidator(current, validator)));
}
function removeValidators(validators, currentValidators) {
    return makeValidatorsArray(currentValidators).filter((validator) => !hasValidator(validators, validator));
}
export class AbstractControl {
    constructor(validators, asyncValidators) {
        Object.defineProperty(this, "_parent", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "_pristine", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: true
        });
        Object.defineProperty(this, "_errors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "_hasOwnPendingAsyncValidator", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
        Object.defineProperty(this, "_asyncValidationSubscription", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_composedValidatorFn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_composedAsyncValidatorFn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_rawValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_rawAsyncValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "asyncVerifyComplete$", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Subject()
        });
        this._rawValidators = validators;
        this._rawAsyncValidators = asyncValidators;
        this._composedValidatorFn = coerceToValidator(this._rawValidators);
        this._composedAsyncValidatorFn = coerceToAsyncValidator(this._rawAsyncValidators);
    }
    get validator() {
        return this._composedValidatorFn;
    }
    set validator(validatorFn) {
        this._rawValidators = this._composedValidatorFn = validatorFn;
    }
    get asyncValidator() {
        return this._composedAsyncValidatorFn;
    }
    set asyncValidator(asyncValidatorFn) {
        this._rawAsyncValidators = this._composedAsyncValidatorFn = asyncValidatorFn;
    }
    get parent() {
        return this._parent;
    }
    get root() {
        let control = this;
        while (control.parent) {
            control = control.parent;
        }
        return control;
    }
    get value() {
        return this._value;
    }
    get status() {
        return this._status;
    }
    get errors() {
        return this._errors;
    }
    get valid() {
        return this._status === VALID;
    }
    get invalid() {
        return this._status === INVALID;
    }
    get pending() {
        return this._status === PENDING;
    }
    get disabled() {
        return this._status === DISABLED;
    }
    get enabled() {
        return this._status !== DISABLED;
    }
    get pristine() {
        return this._pristine;
    }
    get dirty() {
        return !this._pristine;
    }
    clone() {
        return new Proxy(this, {});
    }
    setValidators(validators) {
        this._rawValidators = validators;
        this._composedValidatorFn = coerceToValidator(validators);
    }
    setAsyncValidators(validators) {
        this._rawAsyncValidators = validators;
        this._composedAsyncValidatorFn = coerceToAsyncValidator(validators);
    }
    addValidators(validators) {
        this.setValidators(addValidators(validators, this._rawValidators));
    }
    addAsyncValidators(validators) {
        this.setAsyncValidators(addValidators(validators, this._rawAsyncValidators));
    }
    removeValidators(validators) {
        this.setValidators(removeValidators(validators, this._rawValidators));
    }
    removeAsyncValidators(validators) {
        this.setAsyncValidators(removeValidators(validators, this._rawAsyncValidators));
    }
    hasValidator(validator) {
        return hasValidator(this._rawValidators, validator);
    }
    hasAsyncValidator(validator) {
        return hasValidator(this._rawAsyncValidators, validator);
    }
    clearValidators() {
        this.validator = null;
    }
    clearAsyncValidators() {
        this.asyncValidator = null;
    }
    markAsDirty(onlySelf = false) {
        this._pristine = false;
        if (this._parent && !onlySelf) {
            this._parent.markAsDirty(onlySelf);
        }
    }
    markAsPristine(onlySelf = false) {
        this._pristine = true;
        this._forEachChild((control) => {
            control.markAsPristine(true);
        });
        if (this._parent && !onlySelf) {
            this._parent._updatePristine(onlySelf);
        }
    }
    markAsPending(onlySelf = false) {
        this._status = PENDING;
        if (this._parent && !onlySelf) {
            this._parent.markAsPending(onlySelf);
        }
    }
    disable(onlySelf = false) {
        // If parent has been marked artificially dirty we don't want to re-calculate the
        // parent's dirtiness based on the children.
        const skipPristineCheck = this._parentMarkedDirty(onlySelf);
        this._status = DISABLED;
        this._forEachChild((control) => {
            control.disable(true);
        });
        this._updateAncestors(onlySelf, skipPristineCheck);
    }
    enable(onlySelf = false) {
        // If parent has been marked artificially dirty we don't want to re-calculate the
        // parent's dirtiness based on the children.
        const skipPristineCheck = this._parentMarkedDirty(onlySelf);
        this._status = VALID;
        this._forEachChild((control) => {
            control.enable(true);
        });
        this.updateValueAndValidity(true);
        this._updateAncestors(onlySelf, skipPristineCheck);
    }
    setParent(parent) {
        this._parent = parent;
    }
    setErrors(errors) {
        this._errors = errors;
        this._updateControlsErrors();
    }
    updateValueAndValidity(onlySelf = false) {
        this._updateValue();
        if (this.enabled) {
            this._cancelExistingSubscription();
            this._errors = this._runValidator();
            this._status = this._calculateStatus();
            if (this.status === VALID || this.status === PENDING) {
                this._runAsyncValidator();
            }
        }
        if (this._parent && !onlySelf) {
            this._parent.updateValueAndValidity(onlySelf);
        }
    }
    get(path) {
        return find(this, path, '.');
    }
    getError(errorCode, path) {
        const control = path ? this.get(path) : this;
        return control && control.errors ? control.errors[errorCode] : null;
    }
    hasError(errorCode, path) {
        return !!this.getError(errorCode, path);
    }
    _anyControlsHaveStatus(status) {
        return this._anyControls((control) => control.status === status);
    }
    _anyControlsDirty() {
        return this._anyControls((control) => control.dirty);
    }
    _parentMarkedDirty(onlySelf) {
        return !!(!onlySelf && this._parent && this._parent.dirty && !this._parent._anyControlsDirty());
    }
    _updatePristine(onlySelf = false) {
        this._pristine = !this._anyControlsDirty();
        if (this._parent && !onlySelf) {
            this._parent._updatePristine(onlySelf);
        }
    }
    _updateControlsErrors() {
        this._status = this._calculateStatus();
        if (this._parent) {
            this._parent._updateControlsErrors();
        }
    }
    _calculateStatus() {
        if (this._allControlsDisabled())
            return DISABLED;
        if (this.errors)
            return INVALID;
        if (this._hasOwnPendingAsyncValidator || this._anyControlsHaveStatus(PENDING))
            return PENDING;
        if (this._anyControlsHaveStatus(INVALID))
            return INVALID;
        return VALID;
    }
    _updateAncestors(onlySelf, skipPristineCheck) {
        if (this._parent && !onlySelf) {
            this._parent.updateValueAndValidity(onlySelf);
            if (!skipPristineCheck) {
                this._parent._updatePristine();
            }
        }
    }
    _runValidator() {
        return this.validator ? this.validator(this) : null;
    }
    _cancelExistingSubscription() {
        if (this._asyncValidationSubscription) {
            this._asyncValidationSubscription.unsubscribe();
            this._hasOwnPendingAsyncValidator = false;
        }
    }
    _runAsyncValidator() {
        if (this.asyncValidator) {
            this._status = PENDING;
            this._hasOwnPendingAsyncValidator = true;
            this._asyncValidationSubscription = from(this.asyncValidator(this)).subscribe((errors) => {
                this._hasOwnPendingAsyncValidator = false;
                // This will trigger the recalculation of the validation status, which depends on
                // the state of the asynchronous validation (whether it is in progress or not). So, it is
                // necessary that we have updated the `_hasOwnPendingAsyncValidator` boolean flag first.
                this.setErrors(errors);
                this.asyncVerifyComplete$.next(this);
            });
        }
    }
}
